//v0.1
/*Todo
 
 * Better log function
 * Better statistics
 * WiFi statistics
*/

#include "Arduino.h"
#include "JacobServer.h"




JacobServer::JacobServer()
{
 server[0]='\0';
}

void JacobServer::setServerName(char* serverDNSstring)
{
  strcpy(server, serverDNSstring);
}

#ifdef ARDUINO_ARCH_AVR
void JacobServer::ethernetSetup(byte mac[], byte ip[])
{
  Serial.print("Ethernet Setup...");
  myMAC[0]=mac[0];
  myMAC[1]=mac[1];
  myMAC[2]=mac[2];
  myMAC[3]=mac[3];
  myMAC[4]=mac[4];
  myMAC[5]=mac[5];
  myIP[0]=ip[0];
  myIP[1]=ip[1];
  myIP[2]=ip[2];
  myIP[3]=ip[3];
  
  while (Ethernet.begin(mac) == 0);
  Ethernet.begin(mac,ip);

  Serial.print(" ... ");

  // give the Ethernet shield time to initialize:
  delay(3000);
  Serial.print(" ...Done\n ");
}
void JacobServer::ethernetReconnect()
{
  Serial.print("Ethernet Setup...");
  
  while (Ethernet.begin(myMAC) == 0);
  Ethernet.begin(myMAC,myIP);

  Serial.print(" ... ");

  // give the Ethernet shield time to initialize:
  delay(3000);
  Serial.print(" ...Done\n ");
}
#endif

#if defined ARDUINO_ARCH_ESP8266 || ARDUINO_ARCH_ESP32
void JacobServer::ethernetSetup(const char *SSID, const char *pwd)
{
  ssidListLength=0;
  Serial.print("Connecting to WiFi network :");
  Serial.print(SSID);
  Serial.print("  ");

  strcpy(mySSID,SSID);
  strcpy(myPwd,pwd);
  
  WiFi.mode(WIFI_STA);
  WiFi.begin(mySSID, myPwd);
  
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("WiFi connected");  
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
}
void JacobServer::ethernetSetupMulti( char *SSID[],  char *pwd[], int numberOfSSID)
{
  int i;
  for (i=0;i<numberOfSSID;i++)
  {
     ssidList[i]=SSID[i];
     pwdList[i]=pwd[i];
  }
  ssidListLength=numberOfSSID;

  

  WiFi.mode(WIFI_STA);

  for (i=0;i<ssidListLength;i++)
     wifiMulti.addAP(ssidList[i],pwdList[i]);

  while (wifiMulti.run() != WL_CONNECTED) { // Wait for the Wi-Fi to connect: scan for Wi-Fi networks, and connect to the strongest of the networks above
    delay(500);
    Serial.print('.');
  }   

  Serial.print("WiFi connected to: ");
  Serial.println(WiFi.SSID());
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
}

void JacobServer::ethernetReconnect()
{
	
  if (hardResetAtReconnect==1) ESP.restart();	
	
  Serial.print("Reonnecting to WiFi network.");

  WiFi.mode(WIFI_OFF);
  delay(200);
  if (ssidListLength==0)
  {
	  ethernetSetup(mySSID,myPwd);
  } else {
	  ethernetSetupMulti(ssidList, pwdList, ssidListLength);
  }
}

#endif


void JacobServer::createStoreMessage(char* message, const char* id)
{
  sprintf(message,"id=%s&cmd=silentStore",id);
  return;
}

void JacobServer::appendStoreMessageInt(char* message,  const char* varName,  const int value)
{
  int messageLength;
  messageLength=strlen(message);
  sprintf(message+messageLength,"&%s=%d",varName,value);
  return;
}
void JacobServer::appendStoreMessageLongInt(char* message,  const char* varName,  const long int value)
{
  int messageLength;
  messageLength=strlen(message);
  sprintf(message+messageLength,"&%s=%ld",varName,value);
  return;
}

void JacobServer::appendStoreMessageFloat(char* message,  const char* varName,  const float value)
//Arduino does not handle sprintf for float. This function only outputs integer + 2 decimals
{
  int value1;
  int value2;
  float value3;
  int messageLength;
  if (value>=0)
  {
     value1 = (int) value;
     value2 = (int) (100* (value-value1));
     messageLength=strlen(message);
     sprintf(message+messageLength,"&%s=%d.%d",varName,value1,value2);
  } else {
     value3=-value;
     value1 = (int) value3;
     value2 = (int) (100* (value3-value1));
     messageLength=strlen(message);
     sprintf(message+messageLength,"&%s=-%d.%d",varName,value1,value2);
    
  }
  return;
}

void JacobServer::appendStoreMessageString(char* message,  const char* varName,  const char* value)
{
  int messageLength;
  messageLength=strlen(message);
  sprintf(message+messageLength,"&%s=%s",varName,value);
  return;
}


int JacobServer::pushToServer(char* outbuf) //Return 0 on success, -1 of failure
{
  long int startTime;
  int connectionAttempts=0;
  int result=0;
  
/*  Serial.print("Push to server:");
  Serial.print(server);
  Serial.print(":1419: ");
  Serial.print(outbuf);
  if (existSubscribers==1) 
  {
      Serial.print(" with ");
	  Serial.print(udp.available());
	  Serial.print(" bytes in udp waiting");
  }
  Serial.println();
*/  
  startTime=millis();
  while (connectionAttempts<5)
  {
	  if (result=arduinoClient.connect(server, 1419)) 
	  {
		  connectionAttempts = 10;
	  } else {
		  delay(10*(connectionAttempts+1));
		  Serial.print("JacobServer::pushToServer: retry connection. Error: ");
		  Serial.println(result);
	  }
	  connectionAttempts++;
  }
  if (connectionAttempts>=10) {
    arduinoClient.print(outbuf);
	updateWriteStatistics((int) (millis()-startTime));
  } 
  else {
    Serial.println("JacobServer::pushToServer: Connection to server failed");
	Serial.print("Message:");
	Serial.println(outbuf);
    arduinoClient.stop();
	updateAccessFailure();
    return -1;
  }
  arduinoClient.stop();

  return 0;
}

void JacobServer::registerService(const char* serviceName, const char* servicePrefix, const char* serviceProtocol, const char* id)
{
	char outbuf[512];
	sprintf(outbuf,"id=%s&cmd=registerService&service=%s&prefix=%s&protocolVersion=%s",id,serviceName,servicePrefix,serviceProtocol);
	pushToServer(outbuf);
}



void JacobServer::logToServer(char* outbuf)
{
  long int startTime;
  
  startTime=millis();
  if (arduinoClient.connect(server, 1418)) {
    arduinoClient.print(outbuf);
	updateWriteStatistics((int) (millis()-startTime));
  } 
  else {
    Serial.println("JacobServer::logToServer: Connection to server failed");
    arduinoClient.stop();
	updateAccessFailure();
    return ;
  }
  arduinoClient.stop();
}


void JacobServer::createReadMessage(char* message, const char* id)
{
  sprintf(message,"id=%s&cmd=read",id);
  return;
}

void JacobServer::appendReadMessage(char* message,  const char* varName)
{
  int messageLength;
  messageLength=strlen(message);
  sprintf(message+messageLength,"&%s=0",varName);
  return;
}

int JacobServer::getParameterString(const char* message, const char* varName, char* value, const int maxLen) //Returns 0 on success, -1 on fail
{
  char *c, *c2;
  if ((c=strstr(message,varName))==NULL) return -1;
  c=c+strlen(varName)+1;
  if ((c2=strchr(c,'&'))==NULL) //last parameter
  {
    strncpy(value,c, maxLen);
	value[maxLen-1]='\0';
  } else {
    if ((c2-c)>=maxLen)
	{
       strncpy(value,c,maxLen);
	   value[maxLen-1]='\0';
	} else {
       strncpy(value,c,c2-c);
   	   value[c2-c]='\0';
	}
  }
  return 0;
}

int JacobServer::getParameterInt(const char* message, const char* varName, int* value) //Returns 0 on success, -1 on fail
{
  char stringValue[10];

  if  (getParameterString(message,varName,stringValue,10) ==0)
  {
	  *value = (int)strtol(stringValue,NULL,10);
	  return 0;
  } else
	  return -1;
 }

int JacobServer::getParameterLongInt(const char* message, const char* varName, long int* value) //Returns 0 on success, -1 on fail
{
  char stringValue[10];

  if  (getParameterString(message,varName,stringValue,10) ==0)
  {
	  *value = strtol(stringValue,NULL,10);
	  return 0;
  } else
	  return -1;
 }

 int JacobServer::getParameterFloat(const char* message, const char* varName, float* value) //Returns 0 on success, -1 on fail
{
  char stringValue[10];

  if  (getParameterString(message,varName,stringValue,10) ==0)
  {
	  *value = (float) strtod(stringValue,NULL);
	  return 0;
  } else
	  return -1;
 }
 
 
void JacobServer::appendReadMessageEntries(char* message,  const ParameterEntry* entries, int length)
{
	int i;
	for (i=0; i<length; i++)
	{
		appendReadMessage(message, entries[i].name);
	}
}

int JacobServer::getParameterEntries(char* message, ParameterEntry* entries, int length)
{
	int i;
	int result=0;
	for (i=0; i<length; i++)
	{
		entries[i].readResult=getParameterString(message, entries[i].name, entries[i].value, entries[i].length);
		result=result+entries[i].readResult;
	}
	return result;
}


 
int JacobServer::readFromServerWithFaultReason(char* message, char* inbuf, int maxLength) //Return characters read on success, -1 if error
{
  char c;
  int i;
  unsigned long startT, stopT;
  int result=0;

  startT=millis();	
  if (result=arduinoClient.connect(server, 1419)) {
    arduinoClient.println(message);
	updateWriteStatistics((int) (millis()-startT));
  } 
  else {
    Serial.print("JacobServer::readFromServerWithFaultReason: Connection to server failed: ");
	Serial.println(result);
	Serial.print("Message:");
	Serial.println(message);
	updateAccessFailure();
    arduinoClient.stop();
	strcpy(inbuf,"Connection to server failed");
    return -1;
  }

  startT=millis();
  stopT=millis()+1500L; //Max 1500ms
  i=0;
  c=arduinoClient.read();
  while ((stopT>millis()) && (((int)c)!=10)) 
  {
    if ((c!=-1) && (c!=0)&& (c!=255))
    {
      inbuf[i]=c;
      if (i<(maxLength-1)) i++;
    }
    c=arduinoClient.read();
  }
  inbuf[i]='\0'; //Terminate string. Note that CR/LF is not included in string

  arduinoClient.stop();

  if (stopT<millis()) 
  {
    Serial.println("JacobServer::readFromServer: TIME OUT during read");
	sprintf(inbuf,"TIME OUT during read. Read %d bytes in %ld ms",i,(millis()-startT));
    readFails++;
    return -1;
  }
  updateReadStatistics((int) (millis()-startT));
  
/*  Serial.print("Read:");
  Serial.print(message);
  Serial.print(":");
  Serial.println(inbuf);
*/
  return i;
}

 
int JacobServer::readFromServer(char* message, char* inbuf, int maxLength) //Return characters read on success, -1 if error
{
  char c;
  int i;
  unsigned long startT, stopT;
  int result=0;

  startT=millis();	
  if (result=arduinoClient.connect(server, 1419)) {
    arduinoClient.println(message);
	updateWriteStatistics((int) (millis()-startT));
  } 
  else {
    Serial.print("JacobServer::readFromServer: Connection to server failed: ");
	Serial.println(result);
	Serial.print("Message:");
	Serial.println(message);
	updateAccessFailure();
    arduinoClient.stop();
    return -1;
  }

  startT=millis();
  stopT=millis()+1500L; //Max 1500ms
  i=0;
  c=arduinoClient.read();
  while ((stopT>millis()) && (((int)c)!=10)) 
  {
    if ((c!=-1) && (c!=0)&& (c!=255))
    {
      inbuf[i]=c;
      if (i<(maxLength-1)) i++;
    }
    c=arduinoClient.read();
  }
  inbuf[i]='\0'; //Terminate string. Note that CR/LF is not included in string

  arduinoClient.stop();

  if (stopT<millis()) 
  {
    Serial.println("JacobServer::readFromServer: TIME OUT during read");

    readFails++;
    return -1;
  }
  updateReadStatistics((int) (millis()-startT));

/*  Serial.print("Read:");
  Serial.print(message);
  Serial.print(":");
  Serial.println(inbuf);
*/
  return i;
}

void JacobServer::clearStatistics()
{
	reconnectAttempts=0;
	filteredWriteFails=0;
	readLatencyTot=0L;
	readLatencyMax=0;
	readLatencyMin=1500;
	writeLatencyTot=0L;
	writeLatencyMax=0;
	writeLatencyMin=8000;
	readFails=0L;
	writeFails=0L;
	readCount=0L;
	writeCount=0L;
	readC0=0L;
	writeC0=0L;
	readC1=0L;
	writeC1=0L;
	readC2=0L;
	writeC2=0L;
	readC3=0L;
	writeC3=0L;
	readC4=0L;
	writeC4=0L;
	writeC5=0L;
	writeC6=0L;
	writeC7=0L;
	writeC8=0L;
#ifdef ARDUINO_ARCH_ESP8266
	rssiMax=-200L;
	rssiMin=1000L;
	rssiC1=0L;
	rssiC0=0L;
	rssiC2=0L;
	rssiC3=0L;
	rssiC4=0L;
#endif

}

void JacobServer::updateReadStatistics(int readTime)
{
	readLatencyTot=readLatencyTot+ (long int)readTime;
	readCount++;
	if (readTime>readLatencyMax) readLatencyMax=readTime;
	if (readTime<readLatencyMin) readLatencyMin=readTime;
	if (readTime<20) readC0++;
	if ((readTime>=20) && (readTime<50)) readC1++;
	if ((readTime>=50) && (readTime<100)) readC2++;
	if ((readTime>=100) && (readTime<200)) readC3++;
	if (readTime>=200) readC4++;
}

void JacobServer::updateWriteStatistics(int writeTime)
{
	long int r;
	
	filteredWriteFails--;
	reconnectAttempts=0;
	if (filteredWriteFails<0) filteredWriteFails=0;
	
	writeLatencyTot=writeLatencyTot+ (long int) writeTime;
	writeCount++;
	if (writeTime>writeLatencyMax) writeLatencyMax=writeTime;
	if (writeTime<writeLatencyMin) writeLatencyMin=writeTime;

	if (writeTime<20) writeC0++;
	if ((writeTime>=20) && (writeTime<50)) writeC1++;
	if ((writeTime>=50) && (writeTime<100)) writeC2++;
	if ((writeTime>=100) && (writeTime<200)) writeC3++;
	if ((writeTime>=200) && (writeTime<300)) writeC4++;
	if ((writeTime>=300) && (writeTime<600)) writeC5++;
	if ((writeTime>=600) && (writeTime<2000)) writeC6++;
	if ((writeTime>=2000) && (writeTime<4000)) writeC7++;
	if (writeTime>=4000) writeC8++;

#ifdef ARDUINO_ARCH_ESP8266
	r = WiFi.RSSI();
	if (r<rssiMin) rssiMin=r;
	if (r>rssiMax) rssiMax=r;
	if (r<-100) rssiC0++;
	if ((r>=-100) && (r<-90)) rssiC1++;
	if ((r>=-90) && (r<-55)) rssiC2++;
	if ((r>=-75) && (r<-60)) rssiC3++;
	if (r>=-60)  rssiC4++;
#endif	
	
}

void JacobServer::updateAccessFailure()
{
	long int r;
	char logmsg[100];
	writeFails++;
	filteredWriteFails++;
	if (filteredWriteFails>20) //20 in a row 
	{

#ifdef ARDUINO_ARCH_ESP8266
		r = WiFi.RSSI();
        sprintf(logmsg,"ESP8266 attempting reconnect: previous attempts: %d connectStatus: %d RSSI: %ld", reconnectAttempts, WiFi.status(), r );
#endif
		
#ifdef ARDUINO_ARCH_AVR
		sprintf(logmsg,"Arduino attempting reconnect: previous attempts: %d",reconnectAttempts);
#endif		
		reconnectAttempts++;
		filteredWriteFails=0;
		ethernetReconnect();
		logToServer(logmsg);
	}
		
}


void JacobServer::logStatistics(char* header, long int minimumCount)
{
	char logmsg[300];
	long int readAvg=0L;
	long int writeAvg=0L;
	if (readCount + writeCount < minimumCount) return;
	if (readCount==0) readAvg=0; else readAvg=readLatencyTot/readCount;
	if (writeCount==0) writeAvg=0; else writeAvg=writeLatencyTot/writeCount;
	sprintf(logmsg,"%s (min, max, avg, fails, tot) read (%d, %d, %ld, %ld, %ld) write (%d, %d, %ld, %ld, %ld)",header, readLatencyMin, readLatencyMax, readAvg, readFails, readCount, writeLatencyMin, writeLatencyMax, writeAvg, writeFails, writeCount);
	logToServer(logmsg);
	sprintf(logmsg,"%s  distribution: read (%ld, %ld, %ld, %ld, %ld) write (%ld, %ld, %ld, %ld, %ld, %ld, %ld, %ld, %ld)",header, readC0, readC1, readC2, readC3, readC4, writeC0, writeC1, writeC2, writeC3, writeC4, writeC5, writeC6, writeC7, writeC8);
	logToServer(logmsg);
#ifdef ARDUINO_ARCH_ESP8266
	sprintf(logmsg,"%s  RSSI min: %ld  max : %ld  distribution:(%ld, %ld, %ld, %ld, %ld)",header, rssiMin, rssiMax, rssiC0, rssiC1, rssiC2, rssiC3, rssiC4);
	logToServer(logmsg);
#endif	

	clearStatistics();
}


void JacobServer::startSubscribing()
{
	int result;
	Serial.print("Start Subscribing");
	result=udp.begin(1419);
	Serial.println(result);
	existSubscribers=1;
}

void JacobServer::registerSubscription( char* id,  char* key,  void (*callback)( void*,  int,  int,  char*),  void* ptr,  int arg1,  int arg2)
{

	
 Serial.print("Register subs: ");
 Serial.print(numberOfSubscribers);
 Serial.print(": ");
 Serial.print(key);
 Serial.println(" ");

    if (numberOfSubscribers==0) startSubscribing();
	subscriber[numberOfSubscribers].compareString=key;
	subscriber[numberOfSubscribers].callback=callback;
	subscriber[numberOfSubscribers].arg1 = arg1;
	subscriber[numberOfSubscribers].arg2 = arg2;
	subscriber[numberOfSubscribers].ptr = ptr;
	subscriber[numberOfSubscribers].id = id;
	
	numberOfSubscribers++;
	if (numberOfSubscribers==NUMBEROFCALLBACKENTRIES) numberOfSubscribers--;

    updateSubscriptions(id);	
}

void JacobServer::updateSubscriptions(char* id)
{
	char subscriptionMessage[SUBSCRIPTIONMESSAGESIZE];
	int i;
//    Serial.println("updateSubscriptions()");
	sprintf(subscriptionMessage,"id=%s&cmd=subscribe",id);
	for (i=0;i<numberOfSubscribers; i++)
	{
		if (strcmp(id,subscriber[i].id)==0)
		{
			strcat(subscriptionMessage,"&");
			strcat(subscriptionMessage, subscriber[i].compareString);
		}
	}

//    Serial.println(subscriptionMessage);	
	pushToServer(subscriptionMessage);
}


void JacobServer::checkReceivedPacket()
{
    int n;
	int i;
	if (incheckReceivedPacket==1) return;
	incheckReceivedPacket=1;
//    Serial.println("checkReceivedPacket()");
    if ((n=udp.parsePacket())>0)
    {
      n=udp.read(udpReadBuffer,UDPREADBUFFERSIZE-1);
	  udpReadBuffer[n]=0;
	  Serial.print("Check receive packet:");
	  Serial.println(udpReadBuffer);
	  for (i=0;i<numberOfSubscribers; i++)
		  if (strstr(udpReadBuffer,subscriber[i].compareString)!=NULL) subscriber[i].callback(subscriber[i].ptr, subscriber[i].arg1, subscriber[i].arg2, udpReadBuffer);
    }
	incheckReceivedPacket=0;

}

		