//v0.1

#ifndef JacobServer_h
#define JacobServer_h

#include "Arduino.h"

#ifdef ARDUINO_ARCH_ESP8266
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WiFiUdp.h>
#endif

#ifdef ARDUINO_ARCH_ESP32
#include <WiFi.h>
#include <WiFiMulti.h>
#include <WiFiUdp.h>
#endif

#ifdef ARDUINO_ARCH_AVR
#include <Ethernet.h>
#include <EthernetUdp.h>
#endif

#define NUMBEROFCALLBACKENTRIES 12
#define UDPREADBUFFERSIZE 128
#define SUBSCRIPTIONMESSAGESIZE 512

struct ParameterEntry
{
	char* name;
	char* value;
	int length;
	int readResult;
};

struct SubscribeCallbackEntry
{
	char* id;
	char* compareString;
    void (*callback)( void*,  int,  int,  char*);
	int arg1;
	int arg2;
	void* ptr;
};



class JacobServer {
  public:
#ifdef ARDUINO_ARCH_AVR
    EthernetClient arduinoClient;
	byte myMAC[6];
	byte myIP[4];
#endif
#if defined ARDUINO_ARCH_ESP8266 || ARDUINO_ARCH_ESP32
    WiFiClient arduinoClient;
	char mySSID[50];
	char myPwd[100];
	char *ssidList[10];
	char *pwdList[10];
    int ssidListLength;
	
#endif
    char server[30]; // name address for DNS
	struct SubscribeCallbackEntry subscriber[NUMBEROFCALLBACKENTRIES];
	int numberOfSubscribers=0;
	long int readLatencyTot=0;
	int readLatencyMax=0;
	int readLatencyMin=1500;
	long int writeLatencyTot=0L;
	int writeLatencyMax=0;
	int writeLatencyMin=8000;
	int reconnectAttempts=0;
	int filteredWriteFails=0;
	long int readFails=0L;
	long int writeFails=0L;
	long int readCount=0L;
	long int writeCount=0L;
	long int readC1=0L;
	long int readC0=0L;
	long int readC2=0L;
	long int readC3=0L;
	long int readC4=0L;
	long int writeC0=0L;
	long int writeC1=0L;
	long int writeC2=0L;
	long int writeC3=0L;
	long int writeC4=0L;
	long int writeC5=0L;
	long int writeC6=0L;
	long int writeC7=0L;
	long int writeC8=0L;
	int hardResetAtReconnect=0;
	int existSubscribers=0;
	int incheckReceivedPacket=0;

#if defined ARDUINO_ARCH_ESP8266 || ARDUINO_ARCH_ESP32
	long int rssiMax=-200L;
	long int rssiMin=1000L;
	long int rssiC1=0L;
	long int rssiC0=0L;
	long int rssiC2=0L;
	long int rssiC3=0L;
	long int rssiC4=0L;

#endif


#ifdef ARDUINO_ARCH_AVR
    EthernetUDP udp;
#endif
#if defined ARDUINO_ARCH_ESP8266
	WiFiUDP udp;
	ESP8266WiFiMulti wifiMulti;
#endif
#if defined ARDUINO_ARCH_ESP32
	WiFiUDP udp;
	WiFiMulti wifiMulti;
#endif
    
	char udpReadBuffer[UDPREADBUFFERSIZE];
	
//    JacobServer(char* serverDNSstring);
	JacobServer();
	void setServerName(char* serverDNSstring);
#ifdef ARDUINO_ARCH_AVR
    void ethernetSetup(byte mac[], byte ip[]);
#endif
#if defined ARDUINO_ARCH_ESP8266 || ARDUINO_ARCH_ESP32
    void ethernetSetup(const char *SSID, const char *pwd);
    void ethernetSetupMulti( char *SSID[],  char *pwd[], int numberOfSSID);
#endif
	void ethernetReconnect();
    void createStoreMessage(char* message, const char* id);
    void appendStoreMessageInt(char* message,  const char* varName,  const int value);
    void appendStoreMessageLongInt(char* message,  const char* varName,  const long int value);
    void appendStoreMessageFloat(char* message,  const char* varName,  const float value);
    void appendStoreMessageString(char* message,  const char* varName,  const char* value);
    int pushToServer(char* outbuf); //Return 0 on success
	void registerService(const char* serviceName, const char* servicePrefix, const char* serviceProtocol, const char* id);
    void logToServer(char* outbuf);
    void createReadMessage(char* message, const char* id);
    void appendReadMessage(char* message,  const char* varName);
    void appendReadMessageEntries(char* message,  const ParameterEntry* entries, int length);
	int getParameterEntries(char* message, ParameterEntry* entries, int length); //Return 0 on success, <0 on fail
    int getParameterString(const char* message, const char* varName, char* value, const int maxLen); //Returns 0 on success, -1 on fail
	int getParameterInt(const char* message, const char* varName, int* value); //Returns 0 on success, -1 on fail
	int getParameterLongInt(const char* message, const char* varName, long int* value); //Returns 0 on success, -1 on fail
    int getParameterFloat(const char* message, const char* varName, float* value); //Returns 0 on success, -1 on fail

    int readFromServer(char* message, char* inbuf, int maxLength); //Return characters read on success, -1 if error
    int readFromServerWithFaultReason(char* message, char* inbuf, int maxLength); //Return characters read on success, -1 if error . Fault in inbuf.
	void clearStatistics();
	void updateReadStatistics(int readTime);
	void updateWriteStatistics(int writeTime);	
	void updateAccessFailure();
	void logStatistics(char* header, long int minimumCount);
	
	void startSubscribing();
	void updateSubscriptions(char* id);

	void registerSubscription(char* id,  char* key,  void (*callback)( void*,  int,  int,  char*),  void* ptr,  int arg1,  int arg2);

	
	
	void checkReceivedPacket();

};

#endif

