//version 0.1
//
// Todo: 

// * Add bitmask for tracing in constructor
//
// * Create 30 day timer instead?
// * Better status report
#include "Arduino.h"

#include <JacobTimer.h>


#ifdef UDPTRACING
JacobTimer::JacobTimer(class JacobServer *existingServer, char* clientname, rTraceClass* tracePtr) 
#endif
#ifndef UDPTRACING
JacobTimer::JacobTimer(class JacobServer *existingServer, char* clientname) 
#endif
{
	int i;

#ifdef UDPTRACING
    trace = tracePtr;
#endif

	server=existingServer;
	strcpy(clientName, clientname);

    for (i=0; i<NUMOFTIMERS; i++)
      timers[i].active=0;

}


JacobTimer::JacobTimer(char* servername, char* clientname) 
{
	int i;
	server = new JacobServer;
	server->setServerName(servername);
	strcpy(clientName, clientname);

    for (i=0; i<NUMOFTIMERS; i++)
      timers[i].active=0;

}

#ifdef ARDUINO_ARCH_AVR
JacobTimer::JacobTimer(byte *mac, byte *ip, char* servername, char* clientname)
{
	int i;
	server = new JacobServer;
	server->setServerName(servername);
	strcpy(clientName, clientname);
	server->ethernetSetup(mac, ip);
    for (i=0; i<NUMOFTIMERS; i++)
      timers[i].active=0;

}
#endif


#ifdef ARDUINO_ARCH_ESP8266
JacobTimer::JacobTimer(const char *SSID, const char *pwd, char* servername, char* clientname)
{
	int i;
	server = new JacobServer;
	strcpy(server->server, servername);
	strcpy(clientName, clientname);
	server->ethernetSetup(SSID,pwd);
    for (i=0; i<NUMOFTIMERS; i++)
      timers[i].active=0;

}
#endif

long int JacobTimer::arduinoTime() //Returns arduino time in tenths of seconds. Roll-over after 2500 days
{
  unsigned long newmillis;
  newmillis=millis();
  if (newmillis<lastmillis)
  {
    millisWrapCounter++;
    lastmillis=newmillis;
  }
  newmillis=newmillis/100L;
  
  return (long int) ((42949673L * millisWrapCounter) + newmillis);
}


void JacobTimer::printDate(TimeOfDayStruct* t)
{
  Serial.print(t->year);
  Serial.print("-");
  Serial.print(t->month);
  Serial.print("-");
  Serial.print(t->day);
  Serial.print(",");
  Serial.print(t->hour);
  Serial.print(":");
  Serial.print(t->minute);
  Serial.print(":");
  Serial.print(t->second);
  Serial.print(" (");
  Serial.print(t->ms);
  Serial.print(",");
  Serial.print(t->ds);
  Serial.println(")");
}

void JacobTimer::printDateToString(TimeOfDayStruct* t, char* s)
{
  sprintf(s,"%d-%02d-%02d,%02d:%02d:%02d",t->year,t->month,t->day,t->hour,t->minute,t->second);
}


long int JacobTimer::secondsDiff(TimeOfDayStruct* t2, TimeOfDayStruct* t1) //Returns t2-t1 in s
{
  long int t2day, t1day; //completed days since 0
  long int difference;

  t2day = t2->year*365 + day0ofmonth[t2->month-1] + t2->day;
  t1day = t1->year*365 + day0ofmonth[t1->month-1] + t1->day;

  difference = ((long int)(t2day-t1day))*86400L + ((long int)(t2->hour-t1->hour))*3600L + ((long int)(t2->minute-t1->minute))*60L + ((long int)(t2->second-t1->second));
  return difference;  
}

long int JacobTimer::dsDiff(TimeOfDayStruct* t2, TimeOfDayStruct* t1) //Returns t2-t1 in ds
{
  return t2->ds-t1->ds;  
}



int JacobTimer::storeDate(TimeOfDayStruct* t, char* date, unsigned long ms, long int ds) //returns 0 if ok, -1 if error
{
  if (strlen(date)!=19) return -1;
  
  date[4]='\0';
  date[7]='\0';
  date[10]='\0';
  date[13]='\0';
  date[16]='\0';
  date[19]='\0';
  
  t->year=strtol(date,NULL,10);
  t->month=strtol(date+5,NULL,10);
  t->day=strtol(date+8,NULL,10);
  t->hour=strtol(date+11,NULL,10);
  t->minute=strtol(date+14,NULL,10);
  t->second=strtol(date+17,NULL,10);
  t->ms=ms;
  t->ds=ds;
  return 0;
}

void JacobTimer::calculateError(TimeOfDayStruct* t, TimeOfDayStruct* ref, int *predictionError, long int *dsPerSError)  //Prediction error in 
{
  long int dateDifference;
  long int dsDifference;
  dateDifference = secondsDiff(t, ref); //Seconds between t and ref
  dsDifference = dsDiff(t,ref); //ds between t and ref

  *predictionError = dateDifference - (long int)((dsDifference*1000L)/dsPerS);
  if (dateDifference!=0)
  {
    *dsPerSError = (dsDifference*1000L)/(dateDifference)-dsPerS;
  } else {
    *dsPerSError=0;
  }
  
}

int JacobTimer::readDateFromServer(TimeOfDayStruct* t) //Return 0 on success
{
  char outbuf[400];
  char dateVar[30];
  char c;
  int i;
  unsigned long startT, stopT;

//  Serial.print("Reading from server:");
  

  server->createReadMessage(outbuf,clientName);
  server->appendReadMessage(outbuf,"DATE");

  if(server->readFromServer(outbuf,outbuf,400)==-1) return -1;

  if (server->getParameterString(outbuf, "DATE", dateVar,29)==-1) 
  {
     Serial.print("JacobTimer::readDateFromServer: DATE NOT FOUND:");  
     Serial.println(outbuf);
     return -1;
  } 
  
  if (strlen(dateVar)!=19) 
  {
     Serial.println("JacobTimer::readDateFromServer: DATE TOO SHORT");  
     return -1;
  } 
  
  return storeDate(t,dateVar, millis(), arduinoTime()); //returns 0 if ok, -1 if error
}

void JacobTimer::estimateStartTime(TimeOfDayStruct *t, long int deltaDs)
{
  long int deltaSecond;
  long int deltaMinute;
  long int deltaHour;
  long int deltaDay;

  deltaDay= deltaDs/864000L;
  deltaDs=deltaDs- 864000L*deltaDay;
  deltaHour = deltaDs/36000L;
  deltaDs=deltaDs - 36000L*deltaHour;
  deltaMinute = deltaDs/600L;
  deltaDs=deltaDs - 600L*deltaMinute;
  deltaSecond = deltaDs/10L;

  t->second = t->second - deltaSecond;
  if (t->second < 0)
  {
    t->second=t->second+60;
    t->minute=t->minute-1;
  }
  t->minute= t->minute- deltaMinute;
  if (t->minute < 0)
  {
    t->minute=t->minute+60;
    t->hour=t->hour-1;
  }
  t->hour = t->hour- deltaHour;
  if (t->hour < 0)
  {
    t->hour=t->hour+24;
    t->day=t->day-1;
  }
  t->day = t->day - deltaDay; //OK with negative day
}


int JacobTimer::PLL(struct TimeOfDayStruct *sample)
{
  int i;
  int e;
  long int corrE;
  long int difference;
  int oldSynchState=0;
  struct TimeOfDayStruct state0Sample;
  char logmsg[255];
  long int dsPerSold;
  char logdate[30];

  oldSynchState=synchState;
  dsPerSold=dsPerS;
  
  if (synchState==1) //Synchornized
  {
    calculateError(sample,&lastGoodTime, &e, &corrE);
    if (abs(e) < 30)
    {
//If error is less than 30s, Use sample
//Set sample as new time ref
       dsPerS = dsPerS + corrE/20;
       lastGoodTime=*sample;
	   synchCounter=0;

/*
	unsigned long lastPLLLog=0L;
	unsigned long PLLLogInterval = 36000L; //Every hour
*/
       if (arduinoTime()-lastPLLLog>=PLLLogInterval)
       {
		  printDateToString(sample,logdate);
		  lastPLLLog=arduinoTime();
          sprintf(logmsg, "%s @ %s: Prediction error: %d dsPerSerror: %ld => dsPerS: %ld -> %ld", clientName, logdate,e, corrE, dsPerSold, dsPerS);
          server->logToServer(logmsg);
		  Serial.print("JacobTimer::PLL: ");
		  Serial.println(logmsg);
       }
    } else {
//If error is larger than 30s, try again, max 5 times
      synchCounter++;
	  
      Serial.print("JacobTimer::PLL: sample > 30s off prediction: Testing new sample (");
      Serial.print(synchCounter);
      Serial.print("/5) Received ");
	  printDateToString(sample,logdate);
	  Serial.print(logdate);
	  Serial.print(", expected ");
#ifdef UDPTRACING
    trace->start("PLL: > 30s off predition (");
	trace->appendInt(synchCounter);
	trace->appendString("/5), received");
	trace->appendString(logdate);
	trace->appendString(" expected ");
#endif
	  predictCurrentTime(&state0Sample); //Reusing variable
	  printDateToString(&state0Sample,logdate);
	  Serial.println(logdate);
#ifdef UDPTRACING
	trace->appendString(logdate);
	trace->send(2);
#endif
	  

      if (synchCounter>5) 
      {
        Serial.print("JacobTimer::PLL: Entering sych state 2 - resynch!");
#ifdef UDPTRACING
	    trace->sendTrace(2,"PLL: Entering synch state 2 - resynch");
#endif
        synchCounter=0;
        synchState=2;
        synchAcquisition=1;
      }
    }

  }
  
  //State0 = initial, State2 = resynch
  if ((synchState==0) || (synchState==2))       //Check that 5 consecutive samples are within 20 seconds from the average
  {
    state0Sample=*sample;
    if (synchState==0) lastGoodTime=*sample;
    estimateStartTime(&state0Sample,state0Sample.ds);
    synchAcquisition=1;
    dsPerS=VCODEFAULT;
    startTimeCandidate[synchCounter]=state0Sample;
    synchCounter++;
    if (synchCounter==6) {
      startTime=state0Sample;
      startTime.ds=0;
      startTime.ms=0;
      difference=secondsDiff(&state0Sample,&startTimeCandidate[0]);
      for (i=1;i<=4; i++)
         difference=difference + secondsDiff(&state0Sample,&startTimeCandidate[0]); //In s
      estimateStartTime(&startTime, difference*2);  //*2 since 5 samples in s, and  wanting ds, i.e. really *10/5

      synchState=1;
      synchCounter=0;
      synchAcquisition=0;
      Serial.print("JacobTimer::PLL: 6 samples received, diff: ");
#ifdef UDPTRACING
	  trace->sendTrace(2,"PLL: 6 samples received, diff:");
#endif	  
      for (i=0; i<=5; i++)
      {
#ifdef UDPTRACING
         trace->start("Sample ");
		 trace->appendInt(i);
		 trace->appendString(": ");
		 trace->appendLongInt(secondsDiff(&startTime,&startTimeCandidate[i]));
		 trace->send(2);
#endif
         Serial.print(secondsDiff(&startTime,&startTimeCandidate[i]));
         Serial.print("   ");
         if (abs(secondsDiff(&startTime,&startTimeCandidate[i]))>20) 
         {
          synchState=0;
          synchAcquisition=1;
         }
      }
      if (synchState==1) 
      {
        lastGoodTime=*sample;
        if (oldSynchState==0)
		{

  		    printDateToString(sample,logdate);
            sprintf(logmsg, "%s @ %s: Locked", clientName, logdate);
            server->logToServer(logmsg);
#ifdef UDPTRACING
			trace->sendTrace(2,"PLL Locked");
#endif
            Serial.println(logmsg);
			for (i=0;i<registeredCallbackAtPLLLocked; i++)
				callbackAtPLLLock[i](callbackAtPLLLockData[i]);  //Successful locking - register timers
		}
      } else{
		  Serial.println("FAILED, taking 6 new samples");
#ifdef UDPTRACING
		  trace->sendTrace(2,"PLL failed. Take 6 new samples");
#endif
	  }
    }
  }

  return 0;
    
}




void JacobTimer::predictCurrentTime(TimeOfDayStruct* t)  //Predict time 
{
  long int difference;
  *t=lastGoodTime; //Last sample
  difference = ((arduinoTime() - t->ds)*1000L)/dsPerS;
  t->ds=arduinoTime();

  t->second = t->second + difference;
  t->minute = t->minute + t->second/60;
  t->second=t->second%60;
  t->hour=t->hour + t->minute/60;
  t->minute=t->minute%60;
  t->day=t->day + t->hour/24;
  t->hour=t->hour%24;
}

void JacobTimer::setTimeToday(TimeOfDayStruct* t, char* date)
{
//"08:00:00"
  long int dateDifference;
  char dateBuf[16];

  predictCurrentTime(t);
  stpcpy(dateBuf,date);
  dateBuf[2]='\0';
  dateBuf[5]='\0';
  dateBuf[8]='\0';
  
  t->hour=strtol(dateBuf,NULL,10);
  t->minute=strtol(dateBuf+3,NULL,10);
  t->second=strtol(dateBuf+6,NULL,10);

  dateDifference = secondsDiff(t, &lastGoodTime);
  t->ds = lastGoodTime.ds + (dateDifference*dsPerS)/1000L; 
}

void JacobTimer::setTimeTomorrow(TimeOfDayStruct* t, char* date)
{
  long int dateDifference;
  char dateBuf[16];
//"08:00:00"
  predictCurrentTime(t);
  stpcpy(dateBuf,date);
  t->day=t->day+1;
  dateBuf[2]='\0';
  dateBuf[5]='\0';
  dateBuf[8]='\0';
  
  t->hour=strtol(dateBuf,NULL,10);
  t->minute=strtol(dateBuf+3,NULL,10);
  t->second=strtol(dateBuf+6,NULL,10);

  dateDifference = secondsDiff(t, &lastGoodTime);
  t->ds = lastGoodTime.ds + (dateDifference*dsPerS)/1000L; 
}


int JacobTimer::timerExists(long int ds, void (*callback)(void* ), int originID) //Returns 1 if the timer is already registered
{
  int i;
  for (i=0; i<NUMOFTIMERS; i++)
  {
    if (timers[i].active==1)
    {
      if ((timers[i].timerDs==ds) && (timers[i].callback==callback) && (timers[i].originID==originID)) return 1;
    }
  }
  return 0;
}


int JacobTimer::addTimer(long int ds, void *data, void (*callback)(void* ), int addAllways, int originID) //Return 0 if added, -1 if not
{
  int i, freeTimer;

  if ((addAllways==0) && (ds<arduinoTime())) return -1;  //Don't add timer if it already has expired, unless addAllways == 1
  if (timerExists(ds,callback,originID)) return -1; //Don´t add a timer which is already added
  
  freeTimer=-1;
  for (i=0; i<NUMOFTIMERS; i++)
    if (timers[i].active==0)
      freeTimer=i;
  if (freeTimer==-1) return -1;
  timers[freeTimer].timerDs=ds;
  timers[freeTimer].data=data;
  timers[freeTimer].callback=callback;
  timers[freeTimer].active=1;
  timers[freeTimer].originID=originID;
  return 0;
}

void JacobTimer::removeTimers(int originID) //Return 0 if added, -1 if not
{
   int i;
   for (i=0; i<NUMOFTIMERS; i++)
      if (timers[i].originID==originID)
        timers[i].active=0;
}


void JacobTimer::printDs(const long int ds)
{
  long int hour=0L;
  long int minute=0L;
  long int second=0L;

  if (ds<0)
  {
    Serial.println("No date"); 
    return;
  }
  second = (ds*1000L)/dsPerS;
  hour = second / 3600L;
  second = second % 3600L;
  minute = second / 60L;
  second = second % 60;
  Serial.print(hour);
  Serial.print("h ");
  Serial.print(minute);
  Serial.print("m ");
  Serial.print(second);
  Serial.println("s");
}

void JacobTimer::printDsToString(const long int ds, char* s)
{
  long int hour=0L;
  long int minute=0L;
  long int second=0L;

  if (ds<0)
  {
    strcpy(s,"NegativeNumber");
    return;
  }
  second = (ds*1000L)/dsPerS;
  hour = second / 3600L;
  second = second % 3600L;
  minute = second / 60L;
  second = second % 60;
  sprintf(s, "%02ldh%02ldm%02lds",hour,minute,second);
}


void JacobTimer::testThing()
{
  struct TimeOfDayStruct t;
  predictCurrentTime(&t);
  Serial.print("JacobTimer::testThing: Timer expired: ");
  printDate(&t);
  
}


long int JacobTimer::nextTimer()
{
  int i;
  long int l;
  int found=0;
  for (i=0;i<NUMOFTIMERS;i++)
    if (timers[i].active==1)
    {
      if (found==0)
      {
        found=1;
        l = timers[i].timerDs;
      } else {
        if (l>timers[i].timerDs)
           l=timers[i].timerDs;
      }
    }
   if (found==0) return -1; 
   return l;
}

long int JacobTimer::nextTimerMyID(int id)
{
  int i;
  long int l;
  int found=0;
  for (i=0;i<NUMOFTIMERS;i++)
    if ((timers[i].active==1) && (timers[i].originID==id))
    {
      if (found==0)
      {
        found=1;
        l = timers[i].timerDs;
      } else {
        if (l>timers[i].timerDs)
           l=timers[i].timerDs;
      }
    }
   if (found==0) return -1; 
   return l;
}





void JacobTimer::updatePLL()
{
  struct TimeOfDayStruct t;
  unsigned long timeSincePLL;
  char outbuf[200];
  char tempString[40];
  char tempParamString[60];

  timeSincePLL=arduinoTime()-lastPLLIteration;

  
  if ((PLLactive==1) && ((timeSincePLL>=PLLIterationIntervalLocked) || ((timeSincePLL>=PLLIterationIntervalUnlocked) && (synchAcquisition!=0)) || ((timeSincePLL>=PLLIterationIntervalUnexpected) && (synchCounter>0) && (synchState==1))))
  {
     if (readDateFromServer(&t)==0)
     {
	   lastPLLIteration=arduinoTime();
       PLL(&t);
     } else {
	   lastPLLIteration=lastPLLIteration+100L; //Wait 10s and try again
#ifdef UDPTRACING
 	   trace->sendTrace(2,"PLL failed to read date from server");
#endif
	 }

     if (synchState!=1)
     {
        Serial.print("JacobTimer::updatePLL: SYNCH STATE ");
        Serial.println(synchState);
#ifdef UDPTRACING
		trace->start("PLL synch state: ");
		trace->appendInt(synchState);
		trace->send(2);
#endif
     }
  } 

  if ((synchAcquisition==0) && ((arduinoTime()-lastServerUpdate) > PLLServerUpdateInterval))
  {
	  lastServerUpdate=arduinoTime();
	  server->createStoreMessage(outbuf,clientName);


	  printDateToString(&lastGoodTime,tempString);
	  strcpy(tempParamString,clientName);
	  strcat(tempParamString,".LastSynch");
	  server->appendStoreMessageString(outbuf,tempParamString,tempString);
	  
	  predictCurrentTime(&t);
	  printDateToString(&t,tempString);
	  strcpy(tempParamString,clientName);
	  strcat(tempParamString,".CurrentTime");
	  server->appendStoreMessageString(outbuf,tempParamString,tempString);
	  
	  server->pushToServer(outbuf);  
  }
}


void JacobTimer::checkTimers()
{
  int i;
  int found=0;
  int foundIndex=0;
  for (i=0;i<NUMOFTIMERS;i++)
    if ((timers[i].active==1) && (timers[i].timerDs<=arduinoTime()))
    {
	  if (found==0)
	  {
		  foundIndex=i;
		  found=1;
	  } else {
		  if(timers[i].timerDs < timers[foundIndex].timerDs)   //Found an earlier one
		  {
			  foundIndex=i;
		  }
	  }
    }
  if (found==1) 
  {
	 activeTimer=foundIndex; //Global variable - if callback wants its data...
     timers[foundIndex].active=0;
     timers[foundIndex].callback(timers[foundIndex].data);
	 checkTimers();
  }

}


void JacobTimer::printTimers()
{
	int i;
    for (i=0;i<NUMOFTIMERS;i++)
       if (timers[i].active==1)
       {
         Serial.print("Active timer: ");
         Serial.print(i);
		 Serial.print(" set by id = ");
		 Serial.print(timers[i].originID);
         Serial.print(", expires in ");
         printDs(timers[i].timerDs - arduinoTime());
       } 
}



void JacobTimer::printStatus()
{
  struct TimeOfDayStruct t;
  int i;
  statusCounter--;
  if (statusCounter<0)
  {
    statusCounter=30;

    Serial.println("******************");
    Serial.print("Last sample: ");
    printDate(&lastGoodTime);
    Serial.print("Current time (predicted): ");
    predictCurrentTime(&t);
    printDate(&t);
	printTimers();
    Serial.println("******************");

  }
}


void JacobTimer::registerCallbackAtPLLLock(void (*callback)(void* ), void* data)
{
	if (registeredCallbackAtPLLLocked<NUMBEROFTIMERCLIENTS)
	{
       callbackAtPLLLock[registeredCallbackAtPLLLocked]=callback;
	   callbackAtPLLLockData[registeredCallbackAtPLLLocked] = data;
	   registeredCallbackAtPLLLocked++;
	} else {
		Serial.println("JacobTimer:registerCallbackAtPLLLock: TOO MANY CALLBACK REGISTERED!");
	}
	
}



void JacobTimer::logTimers()
{
  struct TimeOfDayStruct t;
  char logmsg[255];
  char logdate[30];
  char tempTime[30];
  int i;

  predictCurrentTime(&t);
  printDateToString(&t,logdate);
  sprintf(logmsg, "%s @ %s: TIMERS", clientName, logdate);
  server->logToServer(logmsg);

  for (i=0;i<NUMOFTIMERS;i++)
     if (timers[i].active==1)
     {
		printDsToString(timers[i].timerDs - arduinoTime(), tempTime);
        sprintf(logmsg, "%s @ %s: Timer %d expires in", clientName, logdate, i);
		strcat(logmsg, tempTime);
        server->logToServer(logmsg);
     } 
}

bool JacobTimer::interval(long int* previousInstant, long int intervalDs)
{
	if (arduinoTime()>(*previousInstant + intervalDs))
	{
		*previousInstant=arduinoTime();
		return true;
	} 
	return false;
}

