//version 0.1
//
// Todo: 

// * Add bitmask for tracing in constructor
//
// * Create 30 day timer instead?

#ifndef JacobTimer_h
#define JacobTimer_h

#include "Arduino.h"

#include <JacobServer.h>

#define UDPTRACING

#ifdef UDPTRACING
#include <rTrace.h>
#endif


#define VCODEFAULT 10000L
#define LONGMAX 4294967295L
#define LONGHALF 2147483648L
#define NUMOFTIMERS 32
#define PLLLOGINTERVAL 100
#define NUMBEROFTIMERCLIENTS 5


struct TimerStruct {
  long int timerDs = 0L;
  void *data;
  void (*callback)(void* );
  int active = 0;
  int originID=0;
};

struct TimeOfDayStruct {
  long int ds;
  unsigned long ms;
  int year, month, day;
  int hour, minute, second;
};


class JacobTimer {
  public:
    int day0ofmonth[12]={0,31,62,92,123,153,184,215,245,276,306,337};
	class JacobServer *server;


	long int dsPerS=VCODEFAULT;

	int synchState=0;
	int synchAcquisition=1;
	int synchCounter=0;
	int PLLactive=1;

	long int lastPLLIteration=0L;
	long int lastPLLLog=0L;
	long int lastServerUpdate = 0L;
	long int PLLIterationIntervalLocked=600L; //Every 60s
	long int PLLIterationIntervalUnlocked=10L; //Every 1s
	long int PLLIterationIntervalUnexpected=600L; //Every 60s
	long int PLLLogInterval = 1080000L; //Every month
	long int PLLServerUpdateInterval = 20L; //Every 2s

	struct TimeOfDayStruct startTimeCandidate[6];

	struct TimeOfDayStruct startTime;
	struct TimeOfDayStruct lastGoodTime;

	int statusCounter=0;

	struct TimerStruct timers[NUMOFTIMERS+1];
	int activeTimer=0;

	int millisWrapCounter=-1;
	unsigned long lastmillis=LONGHALF;

	char clientName[30];
    void (*callbackAtPLLLock[NUMBEROFTIMERCLIENTS])(void* );
	void *callbackAtPLLLockData[NUMBEROFTIMERCLIENTS];
	int registeredCallbackAtPLLLocked=0;


#ifdef UDPTRACING
    rTraceClass* trace;
#endif


#ifdef UDPTRACING
	JacobTimer(class JacobServer *existingServer, char* clientname,  rTraceClass* tracePtr);
#endif
	
	JacobTimer(class JacobServer *existingServer, char* clientname);
	JacobTimer(char* servername, char* clientname);
#ifdef ARDUINO_ARCH_AVR
	JacobTimer(byte *mac, byte *ip, char* servername, char* clientname);
#endif
#ifdef ARDUINO_ARCH_ESP8266
	JacobTimer(const char *SSID, const char *pwd, char* servername, char* clientname);
#endif
	long int arduinoTime();
	void printDate(TimeOfDayStruct* t);
	void printDateToString(TimeOfDayStruct* t, char* s);
	long int secondsDiff(TimeOfDayStruct* t2, TimeOfDayStruct* t1); //Returns t2-t1 in s
	long int dsDiff(TimeOfDayStruct* t2, TimeOfDayStruct* t1); //Returns t2-t1 in ds
	int storeDate(TimeOfDayStruct* t, char* date, unsigned long ms, long int ds); //returns 0 if ok, -1 if error
	void calculateError(TimeOfDayStruct* t, TimeOfDayStruct* ref, int *predictionError, long int *dsPerSError);  //Prediction error in 
	int readDateFromServer(TimeOfDayStruct* t); //Return 0 on success
	void estimateStartTime(TimeOfDayStruct *t, long int deltaDs);
	int PLL(struct TimeOfDayStruct *sample);
	void predictCurrentTime(TimeOfDayStruct* t);  //Predict time 
	void setTimeToday(TimeOfDayStruct* t, char* date);
	void setTimeTomorrow(TimeOfDayStruct* t, char* date);
	int timerExists(long int ds, void (*callback)(void*), int originID);
	int addTimer(long int ds, void *data, void (*callback)(void*), int addAllways, int originID); //Return 0 if added, -1 if not
	void removeTimers(int originID);
	void printDs(const long int ds);
	void printDsToString(const long int ds, char* s);
	long int nextTimer();
	long int nextTimerMyID(int id);
	void updatePLL();
	void checkTimers();
    void printTimers();
	void printStatus();
	void testThing();
	void registerCallbackAtPLLLock(void (*callback)(void*), void* data);
	void logTimers();
	bool interval(long int* previousInstant, long int intervalDs);


};

#endif

