

#include <rTrace.h>
 

rTraceClass::rTraceClass()
{
	traceName[0]='\0';
	traceWriteBuffer[0]='\0';
	udpReadBuffer[0]='\0';
	traceMask=127;
	traceToSerial=1;
}

   
   
#ifdef ARDUINO_ARCH_AVR
void rTraceClass::connectToNetwork(const byte* mac, const byte* ip)
{
  Serial.print("rTrace: Ethernet Setup...");

  while (Ethernet.begin(mac) == 0);
  Ethernet.begin(mac,ip);

  Serial.print(" ... ");

  // give the Ethernet shield a second to initialize:
  delay(3000);
  Serial.print(" ...Done\n ");
}
#endif

#if defined ARDUINO_ARCH_ESP8266 || ARDUINO_ARCH_ESP32
void rTraceClass::connectToNetwork(const char* ssid, const char* password)
{
  Serial.print("rTrace: Connecting to ");
  Serial.println(ssid);
  
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("");
  Serial.print("rTrace: WiFi connected, received IP: ");  
  Serial.println(WiFi.localIP());

}

void rTraceClass::ethernetSetupMulti( char *SSID[],  char *pwd[], int numberOfSSID)
{
  int i;

  WiFi.mode(WIFI_STA);

  for (i=0;i<numberOfSSID;i++)
     wifiMulti.addAP(SSID[i],pwd[i]);

  while (wifiMulti.run() != WL_CONNECTED) { // Wait for the Wi-Fi to connect: scan for Wi-Fi networks, and connect to the strongest of the networks above
    delay(500);
    Serial.print('.');
  }   

  Serial.print("rTrace: WiFi connected to: ");
  Serial.println(WiFi.SSID());
  Serial.print("rTrace:IP address: ");
  Serial.println(WiFi.localIP());
}

#endif



void rTraceClass::init(const char* traceClientName, const byte* traceServerIP, int udpPortNumber, byte mask, byte copyToSerial)
{
   traceMask = mask;
   traceToSerial = copyToSerial;
   strcpy(traceName, traceClientName);
   memcpy(serverIP, traceServerIP, 4);
   udpPort = udpPortNumber;
   udp.begin(udpPort);
   sendName();
   sendMask(); 
}

void rTraceClass::setMask(byte mask, byte copyToSerial)
{
   traceMask = mask;
   traceToSerial = copyToSerial;
   sendMask(); 
}



void rTraceClass::sendTrace(char traceCause, const char *s)
{
  if ((traceMask & traceCause)==0) return;
  udp.beginPacket(serverIP, udpPort);
  traceCause=traceCause | 128; //Always 1 in msb to avoid a potential \0....
  udp.write((const uint8_t*)&traceCause,1);
  udp.write((const uint8_t*)s, strlen(s));
  udp.endPacket();

  if (traceToSerial==1)
  {
     Serial.print("rTrace: Tracing cause(");
     Serial.print((int)(traceCause & 127));
     Serial.print("): ");
     Serial.println(s);
  }
}


void rTraceClass::start(const char *s)
{
   strcpy(traceWriteBuffer, s);
}

void rTraceClass::appendInt(const int i)
{
  char s[10];
  sprintf(s,"%d",i);
  strcat(traceWriteBuffer, s);
}
void rTraceClass::appendString(char* s)
{
  strcat(traceWriteBuffer, s);
}
void rTraceClass::appendLongInt(const long int i)
{
  char s[10];
  sprintf(s,"%ld",i);
  strcat(traceWriteBuffer, s);
}
void rTraceClass::appendFloat(const float value)
{
  char s[10];
  int value1;
  int value2;
  float value3;
  if (value>=0)
  {
     value1 = (int) value;
     value2 = (int) (100* (value-value1));
     sprintf(s,"%d.%d",value1,value2);
  } else {
     value3=-value;
     value1 = (int) value3;
     value2 = (int) (100* (value3-value1));
     sprintf(s,"-%d.%d",value1,value2); 
  }
  strcat(traceWriteBuffer, s);
}

void rTraceClass::send(char traceCause)
{
  sendTrace(traceCause, traceWriteBuffer);
}

void rTraceClass::sendMask()
{
  byte msg[2] = {1,0};
  msg[1]=traceMask;
  udp.beginPacket(serverIP, udpPort);
  udp.write((const uint8_t*)msg,2);
  udp.endPacket();
  if (traceToSerial==1)
  {
     Serial.print("rTrace: Sent traceMask: ");
     Serial.println(traceMask);
  }
}

void rTraceClass::sendName()
{
  byte msg =2;
  
  udp.beginPacket(serverIP, udpPort);
  udp.write((const uint8_t*)&msg,1);
  udp.write((const uint8_t*)traceName, strlen(traceName));
  udp.endPacket();
  Serial.print("Sent traceName: ");
  Serial.println(traceName);
}


void rTraceClass::checkReceivedPacket()
{
    int n;

    if ((n=udp.parsePacket())>0)
    {
      udp.read(udpReadBuffer,TRACEREADBUFSIZE-1);
      switch(udpReadBuffer[0])
      {
        case 1: //Please ID yourself
           sendName();
           sendMask();
		   sendTrace(1, "Responded to ID request");
           break;
        case 2: //Set trace mask
           traceMask = udpReadBuffer[1];
           sendMask();
           break;
        case 3: //Set trace mask
		   sendTrace(1, "Restart!");
		   delay(100);
#if defined ARDUINO_ARCH_ESP8266 || ARDUINO_ARCH_ESP32
           ESP.restart();
#endif
           break;
        default:
           Serial.print("rTrace: Unknown message from server: ");
           Serial.print((int)udpReadBuffer[0]);
           Serial.print(",");
           Serial.print((int)udpReadBuffer[1]);
           Serial.print(",");
           Serial.print((int)udpReadBuffer[2]);
           Serial.print(",");
           Serial.print((int)udpReadBuffer[3]);
           Serial.print(",");
           Serial.println((int)udpReadBuffer[4]);
      }
    }

}




