#ifndef _RTRACELIB
#define _RTRACELIB

#include "Arduino.h"


#ifdef ARDUINO_ARCH_ESP8266
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WiFiUdp.h>
#endif

#ifdef ARDUINO_ARCH_ESP32
#include <WiFi.h>
#include <WiFiMulti.h>
#include <WiFiUdp.h>
#endif

#ifdef ARDUINO_ARCH_AVR
#include <Ethernet.h>
#include <EthernetUdp.h>
#endif


#define TRACEWRITEBUFSIZE 255 //Longest trace line
#define TRACEREADBUFSIZE 16   //Longest commnad from server
#define TRACENAMESIZE 127     //Longest name of client name   


class rTraceClass {
  public:
#ifdef ARDUINO_ARCH_AVR
    EthernetUDP udp;
#endif
#if defined ARDUINO_ARCH_ESP8266
	WiFiUDP udp;
	ESP8266WiFiMulti wifiMulti;
#endif
#if defined ARDUINO_ARCH_ESP32
	WiFiUDP udp;
	WiFiMulti wifiMulti;
#endif
   char traceMask=127;
   char traceToSerial = 1;
   char traceName[TRACENAMESIZE];
   byte serverIP[4];
   int udpPort;
   char traceWriteBuffer[TRACEWRITEBUFSIZE];
   char udpReadBuffer[TRACEREADBUFSIZE];

#ifdef ARDUINO_ARCH_AVR
    void connectToNetwork(const byte* mac, const byte* ip);
#endif
#ifdef ARDUINO_ARCH_ESP8266
	void connectToNetwork(const char* ssid, const char* password);
    void ethernetSetupMulti( char *SSID[],  char *pwd[], int numberOfSSID);
#endif
#ifdef ARDUINO_ARCH_ESP32
	void connectToNetwork(const char* ssid, const char* password);
    void ethernetSetupMulti( char *SSID[],  char *pwd[], int numberOfSSID);
#endif

   rTraceClass();
   void init(const char* traceClientName, const byte* traceServerIP, int udpPortNumber, byte mask, byte copyToSerial);
   void setMask(byte mask, byte copyToSerial);
   void sendTrace(char traceCause, const char *s);
   void start(const char *s);
   void appendInt(const int i);
   void appendString(char* s);
   void appendLongInt(const long int i);
   void appendFloat(const float value);
   void send(char traceCause);
   void sendMask();
   void sendName();
   void checkReceivedPacket();
   
};
#endif 
   
